// 先准备fields, 钻取的行列fields, 画布的高度与宽度,
import * as _ from "lodash";

export class EchartsGridOptions {
    instance?: any;
    fields: string[];
    columns: {
        fields: string[] | any[],
        type: string,
        splitNumber?: number
    };
    rows: {
        fields: string[] | any[],
        type: string,
        splitNumber?: number
    };
    dataset: any[];
    minCellWidth: number;
    minCellHeight: number;
    headerHeight: number;
    headerWidth: number;
    canvasWidth: number;
    canvasHeight: number;
    seriesOptions: any;
    xAxisOptions?: any;
    yAxisOptions?: any;
}

export function init(options: EchartsGridOptions) {
    // 增加rowKey
    options.dataset = options.dataset.map((d, i) => ({
        d,
        i
    }));

    const bgColor = "#ccc";
    const fontColor = "#000";
    const columnFields = options.columns.fields.slice(0, options.columns.fields.length - 1);
    const rowFields = options.rows.fields.slice(0, options.rows.fields.length - 1);
    let xField: string[] = options.rows.fields.length > 0 ? options.rows.fields[options.rows.fields.length - 1] : [];
    let yField: string[] = options.columns.fields.length > 0 ? options.columns.fields[options.columns.fields.length - 1] : [];
    xField = typeof xField === "string" ? [xField] : xField;
    yField = typeof yField === "string" ? [yField] : yField;
    const columns = getDatum(columnFields, null, 0, options.columns);
    const rows = getDatum(rowFields, null, 0, options.rows);
    const rowsMeta: Array<{
        height: number;
        dataset: any[];
    }> = [];
    const columnsMeta: Array<{
        width: number;
        dataset: any[];
    }> = [];

    const columnsHeight = rows.level * options.headerHeight;
    const rowsWidth = columns.level * options.headerWidth;
    let graphics: any[] = [];
    let grids: any[] = [];
    let series: any[] = [];
    let xAxis: any[] = [];
    let yAxis: any[] = [];
    let singleAxis: any[] = [];
    const scrollBarHeight = 10;
    const scrollBarWidth = 10;
    const scrollHeight = 50;
    const scrollWidth = 50;
    const scrollXPadding = 40;
    let showXScroller = false;
    let showYScroller = false;
    let scrollOffsetX = 0;
    let scrollOffsetY = 0;
    let cellHeight = (options.canvasHeight - columnsHeight) / (columns.leaves * yField.length);
    showYScroller = cellHeight < options.minCellHeight;
    let cellWidth = (options.canvasWidth - rowsWidth) / (rows.leaves * xField.length);
    showXScroller = cellWidth < options.minCellWidth;

    if (showXScroller && !showYScroller) {
        cellHeight = (options.canvasHeight - columnsHeight - scrollHeight - scrollXPadding) / (columns.leaves * yField.length);
    } else if (showYScroller) {
        cellHeight = options.minCellHeight;
    }

    if (!showXScroller && showYScroller) {
        cellWidth = (options.canvasWidth - rowsWidth - scrollWidth) / (rows.leaves * xField.length);
    } else if (showXScroller) {
        cellWidth = options.minCellWidth;
    }

    function renderScroller() {
        let offsetY = columnsHeight + columns.leaves * cellHeight * yField.length + scrollHeight;
        if (options.canvasHeight - scrollHeight < offsetY) {
            offsetY = options.canvasHeight - scrollHeight;
        }
        let gridWidth = rows.leaves * cellWidth * xField.length;
        let width = gridWidth;
        if (options.canvasWidth - rowsWidth - scrollWidth < gridWidth) {
            width = options.canvasWidth - rowsWidth - scrollWidth;
        }
        showXScroller = gridWidth - width > 0;
        if (showXScroller) {
            xScrolls(graphics, width, scrollBarHeight, rowsWidth, offsetY, gridWidth - width + 10);
        }

        let gridHeight = columns.leaves * cellHeight * yField.length;
        let height = gridHeight;
        let screenHeight = options.canvasHeight - columnsHeight - (showXScroller ? scrollHeight : 0) - scrollXPadding;
        if (screenHeight < gridHeight) {
            height = screenHeight;
        }
        if (gridHeight - height > 0) {
            yScrolls(graphics, scrollBarWidth, height, width + rowsWidth, columnsHeight, gridHeight - height);
        }
    }

    function getDatum(fields: string[], keys: { [key: number]: any }, start: number, typeDef: any) {
        const fieldIndex = options.fields.findIndex(field => field === fields[start]);
        start++;
        const datum: {
            keys: string[],
            map: Map<string, any>,
            level?: number,
            leaves: number,
            dataset: any[],
        } = {
            keys: [],
            map: new Map<string, any>(),
            level: 1,
            leaves: 0,
            dataset: options.dataset
        };
        const dataset = options.dataset.filter((di) => {
            const d = di.d;
            let match: boolean = true;
            if (keys) {
                for (let key of _.keys(keys)) {
                    match = d[key] === keys[key];
                    if (!match)
                        break;
                }
            }
            if (match && fieldIndex > -1) {
                if (datum.keys.indexOf(d[fieldIndex]) < 0) {
                    datum.keys.push(d[fieldIndex]);
                }
            }
            return match;
        });
        for (let key of datum.keys) {
            let addKey: any = {};
            addKey[fieldIndex] = key;
            datum.map[key] = getDatum(fields, {...keys, ...addKey}, start, typeDef);
            datum.leaves += datum.map[key].leaves;
        }
        if (datum.keys[0]) {
            datum.level += datum.map[datum.keys[0]].level;
        }
        if (fieldIndex === -1) {
            datum.dataset = dataset;
            datum.leaves = typeDef.type === "dimension" ? dataset.length : typeDef.splitNumber;
            return datum;
        }
        return datum;
    }

    function createCharts(xField: string[], yField: string[], xAxis: any[], yAxis: any[], graphics: any[], grids: any[], series: any[], singleAxis: any[], offsetX: number, offsetY: number) {
        let startY = 0;
        let startX = 0;
        let direction: number = 0;
        if (options.columns.type === "dimension" && options.rows.type === "metric") {
            direction = 1;
        } else if (options.columns.type === options.rows.type) {
            direction = 2;
        }
        columnsMeta.forEach((c, columnIndex) => {
            startY = 0;
            let xFieldIndex = options.fields.findIndex(f => f === xField[columnIndex % xField.length]);
            rowsMeta.forEach((r, rowIndex) => {
                let yFieldIndex = options.fields.findIndex(f => f === yField[rowIndex % yField.length]);
                let dataset: any[] = [];
                if (direction !== 1) {
                    dataset = c.dataset.map(d => {
                        const dd = [...d.d];
                        if (r.dataset.findIndex(cd => cd.i === d.i) < 0) {
                            dd[yFieldIndex] = null;
                        }
                        return dd;
                    });
                } else {
                    dataset = r.dataset.map(d => {
                        const dd = [...d.d];
                        if (c.dataset.findIndex(cd => cd.i === d.i) < 0) {
                            dd[yFieldIndex] = null;
                        }
                        return dd;
                    });
                }

                const gridOption: any = {
                    dataType: "gridItem",
                    x: offsetX + startX,
                    y: offsetY + startY,
                    width: c.width,
                    height: r.height,
                    borderWidth: 0
                };

                const yAxisData = direction === 1 ? r.dataset.map(d => d.d[yFieldIndex]) : null;
                const xAxisData = direction === 0 ? c.dataset.map(d => d.d[xFieldIndex]) : null;
                let seriesData = [];
                switch (direction) {
                    case 1:
                        seriesData = dataset.map(d => d[xFieldIndex]);
                        break;
                    case 2:
                        seriesData = dataset.map(dd => [dd[xFieldIndex], dd[yFieldIndex]]);
                        break;
                    default:
                        seriesData = dataset.map(d => d[yFieldIndex]);
                        break;
                }

                const showYAxis = columnIndex === 0;
                if (showYAxis) {
                    // 植入一个新的图形到grid中
                    grids.push({
                        dataType: "gridYAxis",
                        x: offsetX + startX,
                        y: offsetY + startY,
                        width: c.width,
                        height: r.height,
                        borderWidth: 0
                    });
                    yAxis.push({
                        ...options.yAxisOptions,
                        z: 4,
                        gridIndex: grids.length - 1,
                        data: yAxisData
                    });
                    xAxis.push({
                        gridIndex: grids.length - 1,
                        show: false
                    });
                    series.push({
                        xAxisIndex: xAxis.length - 1,
                        yAxisIndex: yAxis.length - 1,
                        ...options.seriesOptions
                    });
                }

                const showXAxis = rowIndex === rowsMeta.length - 1;
                if (showXAxis) {
                    // 需要判断是否超出边界
                    let xAxisTop = offsetY + startY;
                    if (xAxisTop + r.height + (showXScroller ? scrollHeight : 0) > options.canvasHeight) {
                        xAxisTop = options.canvasHeight - (showXScroller ? scrollHeight : 0) - r.height - scrollXPadding;
                    }
                    grids.push({
                        dataType: "gridXAxis",
                        x: offsetX + startX,
                        y: xAxisTop,
                        z: 7,
                        width: c.width,
                        height: r.height,
                        borderWidth: 0
                    });
                    yAxis.push({
                        show: false,
                        gridIndex: grids.length - 1,
                    });
                    xAxis.push({
                        ...options.xAxisOptions,
                        z: 10,
                        gridIndex: grids.length - 1,
                        data: xAxisData
                    });
                    series.push({
                        ...options.seriesOptions,
                        xAxisIndex: xAxis.length - 1,
                        yAxisIndex: yAxis.length - 1,
                    });
                }

                grids.push(gridOption);
                xAxis.push({
                    ...options.xAxisOptions,
                    gridIndex: grids.length - 1,
                    axisLabel: {show: false},
                    axisTick: {show: false},
                    data: xAxisData
                });

                yAxis.push({
                    ...options.yAxisOptions,
                    gridIndex: grids.length - 1,
                    axisLabel: {show: false},
                    axisTick: {show: false},
                    data: yAxisData
                });

                graphics.push({
                    dataType: "gridItem",
                    type: 'rect',
                    left: offsetX + startX,
                    top: offsetY + startY,
                    shape: {
                        width: c.width,
                        height: r.height
                    },
                    style: {
                        fill: '#fff',
                        stroke: bgColor
                    }
                });
                series.push({
                    xAxisIndex: xAxis.length - 1,
                    yAxisIndex: yAxis.length - 1,
                    data: seriesData,
                    ...options.seriesOptions
                });
                startY += r.height;
            });
            startX += c.width;
        });
    }

    function createColumnHeader(graphics: any[], width: number, height: number, text: string, offsetX: number) {
        graphics.push({
            type: 'rect',
            left: 0,
            top: 0,
            z: 7,
            shape: {width: offsetX, height},
            style: {fill: "#fff"}
        });
        graphics.push({
            type: 'group',
            left: offsetX,
            top: 0,
            width,
            height,
            children: [
                {
                    type: 'rect',
                    left: 0,
                    top: 2,
                    z: 7,
                    shape: {width, height},
                    style: {fill: "#fff"}
                },
                {type: 'text', left: 'center', top: 'center', style: {fill: fontColor, text}}
            ]
        });
    }

    function createRowsHeader(fields: string[], graphics: any[], width: number, height: number, cellWidth: number, offsetY: number) {
        const cellHeight = height - offsetY;
        const childGraphic: any = {
            type: 'group', left: 2, top: offsetY, width, height: cellHeight,
            children: []
        };
        fields.forEach((text, i) => {
            childGraphic.children.push({
                type: 'group',
                left: i * cellWidth,
                top: 0,
                width: cellWidth,
                height: cellHeight,
                children: [
                    {
                        type: 'rect',
                        left: 0,
                        top: 0,
                        z: 8,
                        shape: {width: cellWidth, height: cellHeight},
                        style: {fill: "#fff"}
                    },
                    {
                        type: 'line',
                        left: 0,
                        z: 8,
                        top: cellHeight,
                        shape: {x1: 0, x2: cellWidth},
                        style: {stroke: bgColor}
                    },
                    {type: 'text', left: 'center', top: 'center', z: 10, style: {fill: fontColor, text}}
                ]
            });
        });
        childGraphic.children.push({
            type: 'group',
            left: fields.length * cellWidth,
            top: 0,
            width: cellWidth,
            height: cellHeight,
            children: [
                {
                    type: 'rect',
                    left: 0,
                    top: 0,
                    z: 8,
                    shape: {width: cellWidth - 2, height: cellHeight},
                    style: {fill: "#fff"}
                },
                {
                    type: 'line',
                    left: 0,
                    z: 8,
                    top: cellHeight,
                    shape: {x1: 0, x2: cellWidth},
                    style: {stroke: bgColor}
                }
            ]
        });
        graphics.push(childGraphic);
    }

    function createColumnDimensions(columns: { keys: string[], leaves: number, map: Map<string, any>, dataset: any[] }, graphics: any[], level: number, width: number, height: number, offsetX: number, offsetY: number) {
        const nextLevel = level + 1;
        let nextOffsetX = offsetX;

        columns.keys.forEach((key, i) => {
            let columnWidth = width;
            if (columns.map[key]) {
                columnWidth = columns.map[key].leaves * width * xField.length;
            }
            graphics.push({
                dataType: "header",
                type: 'group',
                left: nextOffsetX,
                top: offsetY + level * height,
                width: columnWidth,
                height,
                children: [
                    {
                        type: 'rect',
                        left: 0,
                        top: 0,
                        z: 7,
                        shape: {width: columnWidth, height},
                        style: {fill: '#fff'}
                    },
                    {
                        type: 'line',
                        left: columnWidth,
                        top: 0,
                        z: 8,
                        shape: {x1: 0, y1: 0, x2: 0, y2: height},
                        style: {stroke: bgColor}
                    },
                    {type: 'text', left: 'center', top: 'center', z: 7, style: {fill: fontColor, text: key}},
                ]
            });

            if (columns.map[key]) {
                createColumnDimensions(columns.map[key] as any, graphics, nextLevel, width, height, nextOffsetX, offsetY);
            } else {
                // 维度集
                columnsMeta.push({
                    width: columnWidth,
                    dataset: columns.dataset
                });
            }
            nextOffsetX += columnWidth;
        });

        // 不存在keys， 直接在右端画横线
        if (columns.keys.length <= 0) {
            xField.forEach((f, i) => {
                graphics.push({
                    dataType: "header",
                    type: 'group',
                    left: nextOffsetX + i * width * columns.leaves,
                    top: level * height,
                    width,
                    height,
                    z: 3,
                    children: [
                        {
                            type: 'rect',
                            left: 0,
                            top: 0,
                            shape: {width: width * columns.leaves, height},
                            style: {fill: '#fff'}
                        },
                        {
                            type: 'line',
                            left: 0,
                            z: 7,
                            top: height,
                            shape: {x1: 0, y1: 0, x2: width * columns.leaves, y2: 0},
                            style: {stroke: bgColor}
                        }
                    ]
                });

                columnsMeta.push({
                    width: columns.leaves * width,
                    dataset: columns.dataset
                });
            });
        }
    }

    function createRowDimensions(columns: { keys: string[], leaves: number, map: Map<string, any>, dataset: any[] }, graphics: any[], level: number, height: number, width: number, offsetX: number, offsetY: number) {
        const nextLevel = level + 1;
        let nextOffsetY = offsetY;
        columns.keys.forEach((key) => {
            let columnHeight = height;
            if (columns.map[key]) {
                columnHeight = columns.map[key].leaves * height * yField.length;
            }
            graphics.push({
                dataType: "leftHeader",
                type: 'group', left: level * width, top: nextOffsetY, width, height: columnHeight,
                children: [
                    {
                        type: 'rect',
                        left: '0',
                        top: 0,
                        z: 3,
                        shape: {width: width + 2, height: columnHeight + 2},
                        style: {fill: '#fff'}
                    },
                    {
                        type: 'line',
                        left: '0',
                        z: 7,
                        top: columnHeight,
                        shape: {x1: 0, y1: 0, x2: width + 2, y2: 0},
                        style: {stroke: bgColor}
                    },
                    {type: 'text', left: 'center', top: 'center', z: 3, style: {fill: fontColor, text: key}}
                ]
            });
            if (columns.map[key]) {
                createRowDimensions(columns.map[key] as any, graphics, nextLevel, height, width, offsetX, nextOffsetY);
            } else {
                // 维度集
                rowsMeta.push({
                    height: columnHeight,
                    dataset: columns.dataset
                });
            }
            nextOffsetY += columnHeight;
        });
        if (columns.keys.length <= 0) {
            for (let i = 0; i < yField.length; i++) {
                graphics.push({
                    dataType: "leftHeader",
                    type: 'group', left: level * width, top: offsetY + i * height * columns.leaves, width: 0, height,
                    children: [
                        {
                            type: 'rect',
                            left: 0,
                            top: 2,
                            z: 4,
                            shape: {width, height: height * columns.leaves},
                            style: {fill: '#fff'}
                        },
                        {
                            type: 'line',
                            left: width,
                            top: 0,
                            z: 5,
                            shape: {x1: 0, y1: 0, x2: 0, y2: height * columns.leaves},
                            style: {stroke: bgColor}
                        },
                        {
                            type: 'line',
                            left: 0,
                            z: 5,
                            top: height * columns.leaves,
                            shape: {x1: 0, y1: 0, x2: width, y2: 0},
                            style: {stroke: bgColor}
                        }
                    ]
                });
                rowsMeta.push({
                    height: height * columns.leaves,
                    dataset: columns.dataset
                });
            }
        }
    }

    function xScrolls(graphics: any[], width: number, height: number, offsetX: number, offsetY: number, left: number) {
        let lastX = 0;
        let scrollerWidth = width * (width / (width + left));
        scrollerWidth = scrollerWidth < 30 ? 30 : scrollerWidth;
        const rate = left / (width - scrollerWidth);
        const leftScroller = {
            type: 'rect',
            left: 5,
            top: -2,
            z: 19,
            shape: {width: scrollerWidth, height: height + 4},
            style: {fill: "#8c8c8c", stroke: bgColor},
            draggable: true,
            ondragend: () => {
                options.instance.clear();
                options.instance.setOption(fitLayout());
            },
            ondrag: (e: any) => {
                let x = e.target.position[0];
                if (x < 0)
                    x = 0;
                if (x > width - scrollerWidth) {
                    x = width - scrollerWidth;
                }
                e.target.position = [x, -2];
                leftScroller.left = x;
                scrollOffsetX = (x - lastX) * rate;
                graphics.filter(g => g.dataType === "header" || g.dataType === "gridItem").forEach(g => {
                    g.left -= scrollOffsetX;
                });
                grids.filter(g => g.dataType === "gridItem" || g.dataType === "gridXAxis").forEach(g => {
                    g.x -= scrollOffsetX;
                });
                lastX = x;
            }
        };

        graphics.push({
            type: 'rect',
            left: 0,
            z: 18,
            top: offsetY - scrollXPadding,
            shape: {width: rowsWidth, height: options.canvasHeight - height + scrollXPadding},
            style: {fill: "#fff"}
        });

        graphics.push({
            type: 'group',
            left: offsetX,
            top: offsetY + scrollHeight / 2 - height / 2,
            width,
            height,
            children: [
                {
                    type: 'rect',
                    left: 0,
                    top: 0,
                    z: 18,
                    shape: {width, height},
                    style: {fill: "transparent", stroke: bgColor}
                },
                leftScroller
            ]
        });
    }

    function yScrolls(graphics: any[], width: number, height: number, offsetX: number, offsetY: number, left: number) {
        let lastY = 0;
        let yScrollHeight = height * (height / (height + left));
        yScrollHeight = yScrollHeight < 30 ? 30 : yScrollHeight;
        const rate = left / (height - yScrollHeight);
        const scroller = {
            type: 'rect',
            left: -2,
            top: 0,
            z: 12,
            shape: {width: width + 4, height: yScrollHeight},
            style: {fill: "#8c8c8c", stroke: bgColor},
            draggable: true,
            ondragend: () => {
                // 对图形进行左偏移
                options.instance.clear();
                options.instance.setOption(fitLayout());
            },
            ondrag: (e: any) => {
                let y = e.target.position[1];
                if (y < 0)
                    y = 0;
                if (y > height - yScrollHeight) {
                    y = height - yScrollHeight;
                }
                e.target.position = [scroller.left, y];
                scroller.top = y;
                graphics.filter(g => g.dataType === "leftHeader" || g.dataType === "gridItem").forEach(g => {
                    g.top -= (y - lastY) * rate;
                });
                grids.filter(g => g.dataType === "gridItem" || g.dataType === "gridYAxis").forEach(g => {
                    g.y -= (y - lastY) * rate;
                });
                lastY = y;
            }
        };

        const xAxisTop = showXScroller ? (options.canvasHeight - scrollXPadding - scrollHeight) : (options.canvasHeight - scrollXPadding);
        graphics.push({
            type: 'rect',
            left: 0,
            z: 7,
            top: xAxisTop,
            shape: {width: options.canvasWidth, height: options.canvasHeight - xAxisTop},
            style: {fill: "#fff"}
        });

        graphics.push({
            type: 'rect',
            left: offsetX,
            z: 12,
            top: 0,
            shape: {width: scrollWidth, height: options.canvasHeight},
            style: {fill: "#fff", stroke: "transparent"}
        });

        graphics.push({
            type: 'group',
            left: offsetX + scrollWidth / 2 - width / 2,
            top: offsetY,
            width,
            height,
            children: [
                {
                    type: 'rect',
                    left: 0,
                    top: 0,
                    z: 12,
                    shape: {width, height},
                    style: {fill: "transparent", stroke: bgColor}
                },
                scroller
            ]
        });
    }

    function fitLayout() {
        let removeIndexs: number[] = [];
        //默认显示范围内的grids, graphic
        const fitGrids = grids.filter((grid, i) => {
            const show = (grid.x + grid.width) >= 0 && (grid.y + grid.height) >= 0 && grid.x <= options.canvasWidth && grid.y <= options.canvasHeight;
            if (!show) {
                removeIndexs.push(i);
            }
            return show;
        });
        const fitSeries = series.filter((v, i) => removeIndexs.indexOf(i) < 0);
        fitSeries.forEach((s, i) => {
            s.yAxisIndex = i;
            s.xAxisIndex = i;
        });
        const fitYAxis = yAxis.filter((v, i) => removeIndexs.indexOf(i) < 0);
        const fitXAxis = xAxis.filter((v, i) => removeIndexs.indexOf(i) < 0);
        fitYAxis.forEach((s, i) => {
            s.gridIndex = i;
        });
        fitXAxis.forEach((s, i) => {
            s.gridIndex = i;
        });
        const fitGraphics = graphics.filter(graphic => {
            return (graphic.left + (graphic.shape ? graphic.shape.width : graphic.width)) >= 0 && graphic.left <= options.canvasWidth && (graphic.top + (graphic.shape ? graphic.shape.height : graphic.height)) >= 0 && graphic.top <= options.canvasHeight;
        });
        return {
            graphic: fitGraphics,
            grid: fitGrids,
            series: fitSeries,
            xAxis: fitXAxis,
            yAxis: fitYAxis
        }
    }

    renderScroller();
    createColumnHeader(graphics, cellWidth * rows.leaves * xField.length, options.headerHeight + 2, options.rows.fields.splice(0, options.rows.fields.length - 1).join(" / "), rowsWidth);
    createColumnDimensions(rows, graphics, 0, cellWidth, options.headerHeight, rowsWidth, options.headerHeight);

    createRowsHeader(options.columns.fields.splice(0, options.columns.fields.length - 1), graphics, rowsWidth, columnsHeight, options.headerWidth, columnsHeight - options.headerHeight);
    createRowDimensions(columns, graphics, 0, cellHeight, options.headerWidth, 0, columnsHeight);
    createCharts(xField, yField, xAxis, yAxis, graphics, grids, series, singleAxis, rowsWidth + 2, columnsHeight + 2);
    return fitLayout();
}
